/**
 * Copyright (c) 2015-2016  Telinc1
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
package com.telinc1.nfls;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import com.telinc1.nfls.field.Field;
import com.telinc1.nfls.field.FieldBatch;
import com.telinc1.nfls.field.FieldBoolean;
import com.telinc1.nfls.field.FieldNull;
import com.telinc1.nfls.field.array.FieldBooleanArray;
import com.telinc1.nfls.field.decimal.FieldDouble;
import com.telinc1.nfls.field.decimal.FieldFloat;
import com.telinc1.nfls.field.decimal.array.FieldDoubleArray;
import com.telinc1.nfls.field.decimal.array.FieldFloatArray;
import com.telinc1.nfls.field.integer.FieldByte;
import com.telinc1.nfls.field.integer.FieldChar;
import com.telinc1.nfls.field.integer.FieldInt;
import com.telinc1.nfls.field.integer.FieldLong;
import com.telinc1.nfls.field.integer.FieldShort;
import com.telinc1.nfls.field.integer.array.FieldByteArray;
import com.telinc1.nfls.field.integer.array.FieldIntArray;
import com.telinc1.nfls.field.integer.array.FieldLongArray;
import com.telinc1.nfls.field.integer.array.FieldShortArray;
import com.telinc1.nfls.field.integer.array.FieldString;

/**
 * Defines a save file in the NFLS format.
 * 
 * @author Telinc1
 */
public class SaveFile {
    /**
     * Contains the first six bytes of an NFLS file.
     * The first four elements are also used for the end of the footer.
     */
    public static final byte[] NFLS = new byte[]{'N', 'F', 'L', 'S', 1, 0};

    /**
     * Contains all the fields in this save file.
     */
    public Map<String, Field> fields;

    private SaveFile(){
        this.fields = new HashMap<>();
    }

    /**
     * Copy constructor: copies all data from the other save file into this one.
     * 
     * @param save the other save file
     */
    public SaveFile(SaveFile save){
        this.fields = save.getFields();
    }

    /**
     * Reads all the data from the file with the given name.
     * 
     * @param filename the name of the save file
     * @see com.telinc1.nfls.SaveFile#readFile(java.io.File)
     */
    public static SaveFile readFile(String filename) throws IOException {
        return SaveFile.readFile(new File(filename));
    }

    /**
     * Reads all the data from the given file.
     * 
     * @param saveFile the save file to read from
     * @see com.telinc1.nfls.SaveFile#readFile(java.io.DataInputStream)
     */
    public static SaveFile readFile(File saveFile) throws IOException {
        return SaveFile.readFile(new DataInputStream(new FileInputStream(saveFile)));
    }

    /**
     * Reads all the data from the given stream.
     * 
     * @param saveFile the stream to read from. It'll be closed by this method.
     * @throws IOException if an I/O error occurs.
     */
    public static SaveFile readFile(DataInputStream saveFile) throws IOException {
        // Get a new save file.
        SaveFile file = new SaveFile();

        // Starting off, verify the header.
        for(int i = 0; i < 6; i++){
            if(saveFile.readByte() != SaveFile.NFLS[i]){
                return null;
            }
        }

        // Extract the data.
        int dataSize = saveFile.readInt();
        short fieldAmount = saveFile.readShort();

        byte[] data = new byte[dataSize];
        saveFile.read(data, 0, dataSize);

        // Verify the data.
        byte[] checksum = new byte[20];
        saveFile.read(checksum, 0, checksum.length);
        byte[] dataHash = SaveFile.calculateChecksum(data);

        if(!Arrays.equals(dataHash, checksum)){
            return null;
        }

        // Verify the footer.
        for(int j = 0; j < 4; j++){
            if(saveFile.readByte() != SaveFile.NFLS[j]){
                return null;
            }
        }

        // The save file looks good. Let's read it!
        // First we need a DataInputStream though.
        DataInputStream dataStream = new DataInputStream(new ByteArrayInputStream(data));

        for(int k = 0; k < fieldAmount; k++){
            Field field = Field.readField(dataStream);

            if(field == null){
                continue;
            }

            file.fields.put(field.getName(), field);
        }

        saveFile.close();
        dataStream.close();

        return file;
    }

    /**
     * Just creates a new blank save file and returns it.
     * 
     * @return the blank save file
     */
    public SaveFile createNewSaveFile(){
        return new SaveFile();
    }

    /**
     * Save the current data to the given filename.
     * 
     * @param filename the name of the file to save to
     * @see com.telinc1.nfls.SaveFile#save(java.io.File)
     */
    public void save(String filename) throws IOException {
        this.save(new File(filename));
    }

    /**
     * Saves the current data to the given file.
     * 
     * @see com.telinc1.nfls.SaveFile#save(java.io.DataOutputStream)
     */
    public void save(File file) throws IOException {
        this.save(new DataOutputStream(new FileOutputStream(file)));
    }

    /**
     * Attempts to save the current data within this save file to the given
     * data stream.
     * 
     * @param stream the stream to write to. It'll be closed by this method.
     * @throws IOException if an I/O error occurs.
     */
    public void save(DataOutputStream stream) throws IOException {
        ByteArrayOutputStream dataByteStream = new ByteArrayOutputStream();
        DataOutputStream dataStream = new DataOutputStream(dataByteStream);

        // Write all the data to an array.
        for(Entry<String, Field> field : this.getFields().entrySet()){
            Field.writeField(field.getValue(), dataStream);
        }

        byte[] data = dataByteStream.toByteArray();

        // Write the first bytes of the header.
        stream.write(SaveFile.NFLS);

        // Write the data amount and size.
        stream.writeInt(data.length);
        stream.writeShort(this.getFields().size());

        // Write the fields.
        stream.write(data);

        // Calculate and write the checksum.
        stream.write(SaveFile.calculateChecksum(data));

        // Write the last bytes.
        stream.write(SaveFile.NFLS, 0, SaveFile.NFLS.length - 2);

        // Close up all the streams.
        dataStream.close();
        dataByteStream.close();
        stream.close();
    }

    /**
     * Calculates the SHA-1 checksum of the given byte array.
     * 
     * @param bytes the data to be hashed
     * @return the hashed data, or a blank array if hashing was unsuccessful
     * @throws java.security.NoSuchAlgorithException if the user is using some
     * kind of outdated or custom JRE which doesn't support basic MessageDigest
     */
    public static byte[] calculateChecksum(byte[] bytes){
        try{
            return MessageDigest.getInstance("SHA-1").digest(bytes);
        }catch(NoSuchAlgorithmException e){
            System.out.println("What the?! What kind of custom JRE which doesn't support MessageDigest are you using?!");
            e.printStackTrace();
        }

        return new byte[20];
    }

    /**
     * Returns the map which represents all the fields in this save file.
     * It can be freely manipulated.
     * 
     * @return the fields in this save file
     */
    public Map<String, Field> getFields(){
        return this.fields;
    }

    /**
     * Gets a class instance of a field by its name.
     * Returns the field, or null if it doesn't exist.
     * 
     * @param name the field's name
     * @return the field which corresponds to the given name
     */
    public Field getField(String name){
        if(this.getFields().containsKey(name)){
            return this.getFields().get(name);
        }

        return null;
    }

    /**
     * Attempts to grab the given null field by its name.
     * Returns the field, or null if it doesn't exist or isn't a null field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldNull getNullField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldNull){
            return (FieldNull)field;
        }

        return null;
    }

    /**
     * Attempts to grab the given batch field by its name.
     * Returns the field, or null if it doesn't exist or isn't a batch field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldBatch getBatchField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldBatch){
            return (FieldBatch)field;
        }

        return null;
    }

    /**
     * Attempts to grab the given boolean field by its name.
     * Returns the field, or null if it doesn't exist or isn't a boolean field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldBoolean getBooleanField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldBoolean){
            return (FieldBoolean)field;
        }

        return null;
    }

    /**
     * Attempts to grab the given byte field by its name.
     * Returns the field, or null if it doesn't exist or isn't a byte field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldByte getByteField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldByte){
            return (FieldByte)field;
        }

        return null;
    }

    /**
     * Attempts to grab the given short field by its name.
     * Returns the field, or null if it doesn't exist or isn't a short field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldShort getShortField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldShort){
            return (FieldShort)field;
        }

        return null;
    }

    /**
     * Attempts to grab the given int field by its name.
     * Returns the field, or null if it doesn't exist or isn't an int field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldInt getIntField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldInt){
            return (FieldInt)field;
        }

        return null;
    }

    /**
     * Attempts to grab the given long field by its name.
     * Returns the field, or null if it doesn't exist or isn't a long field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldLong getLongField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldLong){
            return (FieldLong)field;
        }

        return null;
    }

    /**
     * Attempts to grab the given float field by its name.
     * Returns the field, or null if it doesn't exist or isn't a float field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldFloat getFloatField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldFloat){
            return (FieldFloat)field;
        }

        return null;
    }

    /**
     * Attempts to grab the given double field by its name.
     * Returns the field, or null if it doesn't exist or isn't a double field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldDouble getDoubleField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldDouble){
            return (FieldDouble)field;
        }

        return null;
    }

    /**
     * Attempts to grab the given char field by its name.
     * Returns the field, or null if it doesn't exist or isn't a char field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldChar getCharField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldChar){
            return (FieldChar)field;
        }

        return null;
    }

    /**
     * Attempts to grab the given boolean array field by its name.
     * Returns the field, or null if it doesn't exist or isn't a boolean array field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldBooleanArray getBooleanArrayField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldBooleanArray){
            return (FieldBooleanArray)field;
        }

        return null;
    }
    
    /**
     * Attempts to grab the given byte array field by its name.
     * Returns the field, or null if it doesn't exist or isn't a byte array field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldByteArray getByteArrayField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldByteArray){
            return (FieldByteArray)field;
        }

        return null;
    }
    
    /**
     * Attempts to grab the given short array field by its name.
     * Returns the field, or null if it doesn't exist or isn't a short array field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldShortArray getShortArrayField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldShortArray){
            return (FieldShortArray)field;
        }

        return null;
    }
    
    /**
     * Attempts to grab the given int array field by its name.
     * Returns the field, or null if it doesn't exist or isn't a int array field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldIntArray getIntArrayField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldIntArray){
            return (FieldIntArray)field;
        }

        return null;
    }
    
    /**
     * Attempts to grab the given long array field by its name.
     * Returns the field, or null if it doesn't exist or isn't a long array field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldLongArray getLongArrayField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldLongArray){
            return (FieldLongArray)field;
        }

        return null;
    }
    
    /**
     * Attempts to grab the given float array field by its name.
     * Returns the field, or null if it doesn't exist or isn't a float array field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldFloatArray getFloatArrayField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldFloatArray){
            return (FieldFloatArray)field;
        }

        return null;
    }
    
    /**
     * Attempts to grab the given double array field by its name.
     * Returns the field, or null if it doesn't exist or isn't a double array field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldDoubleArray getDoubleArrayField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldDoubleArray){
            return (FieldDoubleArray)field;
        }

        return null;
    }
    
    /**
     * Attempts to grab the given string field by its name.
     * Returns the field, or null if it doesn't exist or isn't a string field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldString getStringField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldString){
            return (FieldString)field;
        }

        return null;
    }
}
