/**
 * Copyright (c) 2015-2016  Telinc1
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
package com.telinc1.nfls.field;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import com.telinc1.nfls.field.array.FieldBooleanArray;
import com.telinc1.nfls.field.decimal.FieldDouble;
import com.telinc1.nfls.field.decimal.FieldFloat;
import com.telinc1.nfls.field.decimal.array.FieldDoubleArray;
import com.telinc1.nfls.field.integer.FieldByte;
import com.telinc1.nfls.field.integer.FieldInt;
import com.telinc1.nfls.field.integer.FieldLong;
import com.telinc1.nfls.field.integer.FieldShort;
import com.telinc1.nfls.field.integer.array.FieldByteArray;
import com.telinc1.nfls.field.integer.array.FieldIntArray;
import com.telinc1.nfls.field.integer.array.FieldLongArray;
import com.telinc1.nfls.field.integer.array.FieldShortArray;
import com.telinc1.nfls.field.integer.array.FieldString;

/**
 * Defines a field, the base for any way of storing data.
 * 
 * @author Telinc1
 */
public abstract class Field {
    public static final byte FIELD_NULL = 0;
    public static final byte FIELD_BATCH = 1;

    public static final byte FIELD_BOOLEAN = 2;
    public static final byte FIELD_BYTE = 3;
    public static final byte FIELD_SHORT = 4;
    public static final byte FIELD_INT = 5;
    public static final byte FIELD_LONG = 6;
    public static final byte FIELD_FLOAT = 7;
    public static final byte FIELD_DOUBLE = 8;
    public static final byte FIELD_CHAR = 9;

    public static final byte FIELD_BOOLEAN_ARRAY = 10;
    public static final byte FIELD_BYTE_ARRAY = 11;
    public static final byte FIELD_SHORT_ARRAY = 12;
    public static final byte FIELD_INT_ARRAY = 13;
    public static final byte FIELD_LONG_ARRAY = 14;
    public static final byte FIELD_FLOAT_ARRAY = 15;
    public static final byte FIELD_DOUBLE_ARRAY = 16;
    public static final byte FIELD_STRING = 17;

    protected String name;

    public String getName(){
        return this.name;
    }

    public void setName(String name){
        this.name = name;
    }

    /**
     * Converts a type of field to an actual field.
     * 
     * @param type the field's type
     * @return the class representation of the type, or null if it's invalid
     */
    public static Field resolveType(byte type){
        Class<?extends Field> field = null;

        switch(type){
            case Field.FIELD_NULL:
                field = FieldNull.class;
                break;
            case Field.FIELD_BOOLEAN:
                field = FieldBoolean.class;
                break;
            case Field.FIELD_BYTE:
                field = FieldByte.class;
                break;
            case Field.FIELD_SHORT:
                field = FieldShort.class;
                break;
            case Field.FIELD_INT:
                field = FieldInt.class;
                break;
            case Field.FIELD_LONG:
                field = FieldLong.class;
                break;
            case Field.FIELD_FLOAT:
                field = FieldFloat.class;
                break;
            case Field.FIELD_DOUBLE:
                field = FieldDouble.class;
                break;
            case Field.FIELD_BOOLEAN_ARRAY:
                field = FieldBooleanArray.class;
                break;
            case Field.FIELD_BYTE_ARRAY:
                field = FieldByteArray.class;
                break;
            case Field.FIELD_SHORT_ARRAY:
                field = FieldShortArray.class;
                break;
            case Field.FIELD_INT_ARRAY:
                field = FieldIntArray.class;
                break;
            case Field.FIELD_LONG_ARRAY:
                field = FieldLongArray.class;
                break;
            case Field.FIELD_FLOAT_ARRAY:
                field = FieldByteArray.class;
                break;
            case Field.FIELD_DOUBLE_ARRAY:
                field = FieldDoubleArray.class;
                break;
            case Field.FIELD_STRING:    
                field = FieldString.class;
                break;
        }

        if(field != null){
            try{
                return field.newInstance();
            }catch(ReflectiveOperationException e){
                e.printStackTrace();
            }
        }

        return null;
    }

    /**
     * Reads the next field in the given stream.
     * 
     * @param stream the stream to read from
     * @return the read field, or null if there's no valid field
     */
    public static Field readField(DataInputStream stream) throws IOException {
        short nameLength = stream.readShort();
        StringBuilder builder = new StringBuilder();

        for(int i = 0; i < nameLength; i++){
            builder.append((char)stream.readByte());
        }

        Field field = Field.resolveType(stream.readByte());

        if(field == null){
            return null;
        }

        field.setName(builder.toString());
        field.readFromStream(stream);
        return field;
    }

    /**
     * Writes the given field to the given stream.
     * 
     * @param field the field to write
     * @param stream the stream to write to
     */
    public static void writeField(Field field, DataOutputStream stream) throws IOException {
        stream.writeShort(field.getName().length());
        stream.writeBytes(field.getName());
        stream.writeByte(field.getType());
        field.writeToStream(stream);
    }

    /**
     * Retrieves the type of field.
     * Should adhere to the NFLS v.1 specification.
     * 
     * @return the field type
     */
    public abstract byte getType();

    /**
     * Reads the data of the field from the stored stream.
     * 
     * @param stream the stream to read from
     * @throws IOException {@link java.io.DataInputStream}
     */
    public abstract void readFromStream(DataInputStream stream) throws IOException;

    /**
     * Writes the data stored in this field to the given output stream.
     * 
     * @param stream the stream to write to
     * @return the compiled bytes of the field
     * @throws java.io.IOException {@link java.io.DataOutputStream}
     */
    public abstract void writeToStream(DataOutputStream stream) throws IOException;
}
