/**
 * Copyright (c) 2015-2016  Telinc1
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
package com.telinc1.nfls.field;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import com.telinc1.nfls.field.array.FieldBooleanArray;
import com.telinc1.nfls.field.decimal.FieldDouble;
import com.telinc1.nfls.field.decimal.FieldFloat;
import com.telinc1.nfls.field.decimal.array.FieldDoubleArray;
import com.telinc1.nfls.field.decimal.array.FieldFloatArray;
import com.telinc1.nfls.field.integer.FieldByte;
import com.telinc1.nfls.field.integer.FieldChar;
import com.telinc1.nfls.field.integer.FieldInt;
import com.telinc1.nfls.field.integer.FieldLong;
import com.telinc1.nfls.field.integer.FieldShort;
import com.telinc1.nfls.field.integer.array.FieldByteArray;
import com.telinc1.nfls.field.integer.array.FieldIntArray;
import com.telinc1.nfls.field.integer.array.FieldLongArray;
import com.telinc1.nfls.field.integer.array.FieldShortArray;
import com.telinc1.nfls.field.integer.array.FieldString;

public class FieldBatch extends Field {
    private Map<String, Field> fields;

    public FieldBatch(){
        this(new HashMap<String, Field>());
    }

    public FieldBatch(Map<String, Field> fields){
        this.setFields(fields);
    }

    public FieldBatch(Field[] fields){
        this();

        for(Field field : fields){
            this.getFields().put(field.getName(), field);
        }
    }

    public Map<String, Field> getFields(){
        return this.fields;
    }

    public void setFields(Map<String, Field> fields){
        this.fields = fields;
    }

    /* I would have done this with an abstract class if Java supported extending
     * multiple classes, but no.
     * Because an animal is either a human or a male, but never both.
     */
    /**
     * Gets a class instance of a field by its name.
     * Returns the field, or null if it doesn't exist.
     * 
     * @param name the field's name
     * @return the field which corresponds to the given name
     */
    public Field getField(String name){
        if(this.getFields().containsKey(name)){
            return this.getFields().get(name);
        }

        return null;
    }

    /**
     * Attempts to grab the given null field by its name.
     * Returns the field, or null if it doesn't exist or isn't a null field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldNull getNullField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldNull){
            return (FieldNull)field;
        }

        return null;
    }

    /**
     * Attempts to grab the given batch field by its name.
     * Returns the field, or null if it doesn't exist or isn't a batch field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldBatch getBatchField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldBatch){
            return (FieldBatch)field;
        }

        return null;
    }

    /**
     * Attempts to grab the given boolean field by its name.
     * Returns the field, or null if it doesn't exist or isn't a boolean field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldBoolean getBooleanField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldBoolean){
            return (FieldBoolean)field;
        }

        return null;
    }

    /**
     * Attempts to grab the given byte field by its name.
     * Returns the field, or null if it doesn't exist or isn't a byte field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldByte getByteField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldByte){
            return (FieldByte)field;
        }

        return null;
    }

    /**
     * Attempts to grab the given short field by its name.
     * Returns the field, or null if it doesn't exist or isn't a short field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldShort getShortField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldShort){
            return (FieldShort)field;
        }

        return null;
    }

    /**
     * Attempts to grab the given int field by its name.
     * Returns the field, or null if it doesn't exist or isn't an int field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldInt getIntField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldInt){
            return (FieldInt)field;
        }

        return null;
    }

    /**
     * Attempts to grab the given long field by its name.
     * Returns the field, or null if it doesn't exist or isn't a long field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldLong getLongField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldLong){
            return (FieldLong)field;
        }

        return null;
    }

    /**
     * Attempts to grab the given float field by its name.
     * Returns the field, or null if it doesn't exist or isn't a float field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldFloat getFloatField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldFloat){
            return (FieldFloat)field;
        }

        return null;
    }

    /**
     * Attempts to grab the given double field by its name.
     * Returns the field, or null if it doesn't exist or isn't a double field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldDouble getDoubleField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldDouble){
            return (FieldDouble)field;
        }

        return null;
    }

    /**
     * Attempts to grab the given char field by its name.
     * Returns the field, or null if it doesn't exist or isn't a char field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldChar getCharField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldChar){
            return (FieldChar)field;
        }

        return null;
    }

    /**
     * Attempts to grab the given boolean array field by its name.
     * Returns the field, or null if it doesn't exist or isn't a boolean array field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldBooleanArray getBooleanArrayField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldBooleanArray){
            return (FieldBooleanArray)field;
        }

        return null;
    }
    
    /**
     * Attempts to grab the given byte array field by its name.
     * Returns the field, or null if it doesn't exist or isn't a byte array field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldByteArray getByteArrayField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldByteArray){
            return (FieldByteArray)field;
        }

        return null;
    }
    
    /**
     * Attempts to grab the given short array field by its name.
     * Returns the field, or null if it doesn't exist or isn't a short array field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldShortArray getShortArrayField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldShortArray){
            return (FieldShortArray)field;
        }

        return null;
    }
    
    /**
     * Attempts to grab the given int array field by its name.
     * Returns the field, or null if it doesn't exist or isn't a int array field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldIntArray getIntArrayField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldIntArray){
            return (FieldIntArray)field;
        }

        return null;
    }
    
    /**
     * Attempts to grab the given long array field by its name.
     * Returns the field, or null if it doesn't exist or isn't a long array field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldLongArray getLongArrayField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldLongArray){
            return (FieldLongArray)field;
        }

        return null;
    }
    
    /**
     * Attempts to grab the given float array field by its name.
     * Returns the field, or null if it doesn't exist or isn't a float array field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldFloatArray getFloatArrayField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldFloatArray){
            return (FieldFloatArray)field;
        }

        return null;
    }
    
    /**
     * Attempts to grab the given double array field by its name.
     * Returns the field, or null if it doesn't exist or isn't a double array field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldDoubleArray getDoubleArrayField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldDoubleArray){
            return (FieldDoubleArray)field;
        }

        return null;
    }
    
    /**
     * Attempts to grab the given string field by its name.
     * Returns the field, or null if it doesn't exist or isn't a string field.
     * 
     * @param name the name of the field
     * @return the field, or null if it doesn't exist
     */
    public FieldString getStringField(String name){
        Field field = this.getField(name);

        if(field != null && field instanceof FieldString){
            return (FieldString)field;
        }

        return null;
    }

    @Override
    public byte getType(){
        return Field.FIELD_BATCH;
    }

    @Override
    public void writeToStream(DataOutputStream stream) throws IOException {
        stream.writeShort(this.getFields().size());

        for(Entry<String, Field> field : this.getFields().entrySet()){
            field.getValue().writeToStream(stream);
        }
    }

    @Override
    public void readFromStream(DataInputStream stream) throws IOException {
        Map<String, Field> fields = new HashMap<>();
        short length = stream.readShort();

        for(int i = 0; i < length; i++){
            Field field = Field.readField(stream);
            fields.put(field.getName(), field);
        }

        this.setFields(fields);
    }
}
